/*-
  * SPDX-License-Identifier: BSD-3-Clause
  *
  * Copyright (c) 1992, 1993
  *      The Regents of the University of California.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */

/*
  qsort_r.c from freebsd source code, original file:

  https://www.leidinger.net/FreeBSD/dox/libkern/html/d7/da4/qsort_8c_source.html
*/


typedef int              cmp_t(void *, const void *, const void *);
static inline char      *med3(char *, char *, char *, cmp_t *, void *);
static inline void       swapfunc(char *, char *, size_t, int, int);

#define swapcode(TYPE, parmi, parmj, n) {               \
        size_t i = (n) / sizeof (TYPE);                 \
        TYPE *pi = (TYPE *) (parmi);            \
        TYPE *pj = (TYPE *) (parmj);            \
        do {                                            \
                TYPE    t = *pi;                \
                *pi++ = *pj;                            \
                *pj++ = t;                              \
        } while (--i > 0);                              \
}

#define SWAPINIT(TYPE, a, es) swaptype_ ## TYPE =       \
        ((char *)a - (char *)0) % sizeof(TYPE) ||       \
        es % sizeof(TYPE) ? 2 : es == sizeof(TYPE) ? 0 : 1;

static inline void
swapfunc(char *a, char *b, size_t n, int swaptype_long, int swaptype_int)
{
        if (swaptype_long <= 1)
                swapcode(long, a, b, n)
        else if (swaptype_int <= 1)
                swapcode(int, a, b, n)
        else
                swapcode(char, a, b, n)
}

#define swap(a, b)                                      \
        if (swaptype_long == 0) {                       \
                long t = *(long *)(a);                  \
                *(long *)(a) = *(long *)(b);            \
                *(long *)(b) = t;                       \
        } else if (swaptype_int == 0) {                 \
                int t = *(int *)(a);                    \
                *(int *)(a) = *(int *)(b);              \
                *(int *)(b) = t;                        \
        } else                                          \
                swapfunc(a, b, es, swaptype_long, swaptype_int)

#define vecswap(a, b, n)                                \
        if ((n) > 0) swapfunc(a, b, n, swaptype_long, swaptype_int)

#define CMP(t, x, y) (cmp((t), (x), (y)))

static inline char *
med3(char *a, char *b, char *c, cmp_t *cmp, void *thunk)
{
        return CMP(thunk, a, b) < 0 ?
               (CMP(thunk, b, c) < 0 ? b : (CMP(thunk, a, c) < 0 ? c : a ))
              :(CMP(thunk, b, c) > 0 ? b : (CMP(thunk, a, c) < 0 ? a : c ));
}

void
json_create_qsort_r(void *a, size_t n, size_t es, void *thunk, cmp_t *cmp)
{
        char *pa, *pb, *pc, *pd, *pl, *pm, *pn;
        size_t d1, d2;
        int cmp_result;
        int swaptype_long, swaptype_int, swap_cnt;

loop:   SWAPINIT(long, a, es);
        SWAPINIT(int, a, es);
        swap_cnt = 0;
        if (n < 7) {
                for (pm = (char *)a + es; pm < (char *)a + n * es; pm += es)
                        for (pl = pm; 
                             pl > (char *)a && CMP(thunk, pl - es, pl) > 0;
                             pl -= es)
                                swap(pl, pl - es);
                return;
        }
        pm = (char *)a + (n / 2) * es;
        if (n > 7) {
                pl = a;
                pn = (char *)a + (n - 1) * es;
                if (n > 40) {
                        size_t d = (n / 8) * es;

                        pl = med3(pl, pl + d, pl + 2 * d, cmp, thunk);
                        pm = med3(pm - d, pm, pm + d, cmp, thunk);
                        pn = med3(pn - 2 * d, pn - d, pn, cmp, thunk);
                }
                pm = med3(pl, pm, pn, cmp, thunk);
        }
        swap(a, pm);
        pa = pb = (char *)a + es;

        pc = pd = (char *)a + (n - 1) * es;
        for (;;) {
                while (pb <= pc && (cmp_result = CMP(thunk, pb, a)) <= 0) {
                        if (cmp_result == 0) {
                                swap_cnt = 1;
                                swap(pa, pb);
                                pa += es;
                        }
                        pb += es;
                }
                while (pb <= pc && (cmp_result = CMP(thunk, pc, a)) >= 0) {
                        if (cmp_result == 0) {
                                swap_cnt = 1;
                                swap(pc, pd);
                                pd -= es;
                        }
                        pc -= es;
                }
                if (pb > pc)
                        break;
                swap(pb, pc);
                swap_cnt = 1;
                pb += es;
                pc -= es;
        }
        if (swap_cnt == 0) {  /* Switch to insertion sort */
                for (pm = (char *)a + es; pm < (char *)a + n * es; pm += es)
                        for (pl = pm; 
                             pl > (char *)a && CMP(thunk, pl - es, pl) > 0;
                             pl -= es)
                                swap(pl, pl - es);
                return;
        }

        pn = (char *)a + n * es;
        d1 = MIN(pa - (char *)a, pb - pa);
        vecswap(a, pb - d1, d1);
        d1 = MIN(pd - pc, pn - pd - es);
        vecswap(pb, pn - d1, d1);

        d1 = pb - pa;
        d2 = pd - pc;
        if (d1 <= d2) {
                /* Recurse on left partition, then iterate on right partition */
                if (d1 > es) {
                        json_create_qsort_r(a, d1 / es, es, thunk, cmp);
                }
                if (d2 > es) {
                        /* Iterate rather than recurse to save stack space */
                        /* qsort(pn - d2, d2 / es, es, cmp); */
                        a = pn - d2;
                        n = d2 / es;
                        goto loop;
                }
        } else {
                /* Recurse on right partition, then iterate on left partition */
                if (d2 > es) {
                        json_create_qsort_r(pn - d2, d2 / es, es, thunk, cmp);
                }
                if (d1 > es) {
                        /* Iterate rather than recurse to save stack space */
                        /* qsort(a, d1 / es, es, cmp); */
                        n = d1 / es;
                        goto loop;
                }
        }
}

#undef CMP
#undef swap
#undef swapcode
#undef SWAPINIT
#undef vecswap
